@@ -16,12 +16,16 @@ import (
 // A FileSet is the in-memory representation of a
 // parsed file.
 type FileSet struct {
-	Package    string              // package name
-	Specs      map[string]ast.Expr // type specs in file
-	Identities map[string]gen.Elem // processed from specs
-	Directives []string            // raw preprocessor directives
-	Imports    []*ast.ImportSpec   // imports
-	tagName    string              // tag to read field names from
+	Package       string              // package name
+	Specs         map[string]ast.Expr // type specs in file
+	Identities    map[string]gen.Elem // processed from specs
+	Directives    []string            // raw preprocessor directives
+	Imports       []*ast.ImportSpec   // imports
+	CompactFloats bool                // Use smaller floats when feasible
+	ClearOmitted  bool                // Set omitted fields to zero value
+	NewTime       bool                // Set to use -1 extension for time.Time
+	tagName       string              // tag to read field names from
+	pointerRcv    bool                // generate with pointer receivers.
 }
 
 // File parses a file at the relative path
@@ -199,6 +203,7 @@ parse:
 			popstate()
 			continue parse
 		}
+		el.AlwaysPtr(&f.pointerRcv)
 		// push unresolved identities into
 		// the graph of links and resolve after
 		// we've handled every possible named type.
@@ -267,6 +272,9 @@ loop:
 			warnf("empty directive: %q\n", d)
 		}
 	}
+	p.CompactFloats = f.CompactFloats
+	p.ClearOmitted = f.ClearOmitted
+	p.NewTime = f.NewTime
 }
 
 func (f *FileSet) PrintTo(p *gen.Printer) error {
@@ -524,7 +532,7 @@ func (fs *FileSet) parseExpr(e ast.Expr) gen.Elem {
 	case *ast.Ident:
 		b := gen.Ident(e.Name)
 
-		// work to resove this expression
+		// work to resolve this expression
 		// can be done later, once we've resolved
 		// everything else.
 		if b.Value == gen.IDENT {
